"""
molar平台导入导出实体
"""

from enum import Enum
from dataclasses import dataclass, field

from .d3_entity import Pose


OSS_HTTP_ROOT = r"https://molar-app-prod-v5.oss-cn-hangzhou.aliyuncs.com/"


@dataclass
class CameraConfig:
    """
    相机配置
    """

    name: str = field(default="")
    extrinsic: list[list[float]] = field(default_factory=list)
    intrinsic: list[float] = field(default_factory=list)
    distortion: dict[str, float] = field(default_factory=dict)
    fusionModel: str = field(default="")


@dataclass
class CameraConfigV3:
    name: str = field(default=None)
    extrinsic: list = field(default_factory=list)
    intrinsic: list = field(default_factory=list)
    distortion: dict = field(default=None)
    projectionModel: str = field(default=None)


@dataclass
class Location:
    name: str = field(default=None)
    urls: list[str] = field(default=None)
    posMatrix: list[float] = field(default=None)


@dataclass
class LabelConfig:
    """
    标签配置
    """

    key: str = field(default=None)
    color: str = field(default=None)
    name: str = field(default=None)
    draw_type: str = field(default=None)
    alias: str = field(default=None)


# 支持平台任务类型
class PlatTaskType(Enum):
    IAT = "iat"
    PCAT_D23 = "pcat_d23"


@dataclass
class TaskConfig:
    id: str
    type: str = field(default=None)
    camera_num: int = field(default=1)
    labels: dict[str, Label] = field(default_factory=dict)
    origin_config: dict = field(default_factory=dict)
    origin_task: dict = field(default_factory=dict)


@dataclass
class Annotation:
    uuid: str = field(default=None)
    id: int = field(default=None)
    lens_index: int = field(default=0)
    frame_index: int = field(default=-1)
    draw_type: str = field(default=None)
    points: list = field(default_factory=list)
    attributes: dict = field(default_factory=dict)
    group: int = field(default=-1)
    alias: str = field(default=None)
    p_label_id_map: list[int] = field(default_factory=list)
    connection: int = field(default=-1)


@dataclass
class Frame:
    url: str = field(default=None)
    width: int = field(default=1024)
    height: int = field(default=1024)
    frame_index: int = field(default=-1)
    lens_index: int = field(default=0)
    annotations: dict[str, Annotation] = field(default_factory=dict)
    urls: list[str] = field(default_factory=list)
    pose: Pose = field(default=None)
    camera_config: list[CameraConfig] = field(default=None)


@dataclass
class Item:
    uid: str
    batch_uid: str
    task_config: TaskConfig
    frames: dict = field(default_factory=dict)
    origin_data: dict = field(default_factory=dict)


@dataclass
class D2Info:

    @dataclass
    class Size:
        width: int
        height: int

    url: list[str] = field(default_factory=list)
    size: list[Size] = field(default=None)


@dataclass
class Lod:
    baseUrl: str = field(default=None)
    loadUrl: str = "metadata.json"
    offset: list[float] = field(default=None)


@dataclass
class D3Info:
    pcdUrl: list[str]
    imgUrls: list[str] = field(default=None)
    cameraConfigs: list[list[CameraConfig]] = field(default=None)
    locations: list[Location] = field(default=None)


@dataclass
class D4Info:
    pcdUrl: list[str]
    cameraConfigs: list[list[CameraConfig]] = field(default=None)
    locations: list[Location] = field(default=None)


@dataclass
class ImportInfo:
    info: any
    preData: list = field(default_factory=list)



"""
molar http 客户端
"""

import random
import socket
import ssl
from dataclasses import dataclass, field
from enum import Enum

import requests
from joblib import Parallel, delayed
from loguru import logger
from requests.adapters import HTTPAdapter, PoolManager

from fyuneru.http_utils import find_labels, get_item_info, get_task_info


class MolarDomain(Enum):
    """
    域名
    """

    CN = "https://app.molardata.com"
    OTHER = "https://app.abaka.ai"


class SSLAdapter(HTTPAdapter):
    """AI生成的 SSL 适配"""

    def __init__(self, server_hostname: str, **kwargs):
        self.server_hostname = server_hostname
        super().__init__(**kwargs)

    def init_poolmanager(self, *args, **kwargs):
        context = ssl.create_default_context()
        kwargs["ssl_context"] = context
        # 手动设置目标 SNI 域名
        self.poolmanager = PoolManager(
            *args,
            server_hostname=self.server_hostname,  # 关键点
            assert_hostname=self.server_hostname,  # 确保证书匹配
            **kwargs,
        )


@dataclass
class MolarClient:
    """
    molar http 客户端
    """

    token: str
    domain: str = field(default=MolarDomain.CN.value)

    __required_structure = {
        "taskId": None,
        "exportMetadata": {"match": {"itemIds": None}},
    }

    def __to_export_info(self, export_config: dict) -> "MolarClient.ExportInfo":
        """
        将导出配置转换为导出信息
        """
        metadata = export_config.get("exportMetadata", {}) or export_config.get(
            "metadata", {}
        )
        return MolarClient.ExportInfo(
            origin_data=export_config,
            task_uid=export_config["taskId"],
            item_ids=metadata["match"].get("itemIds", []),
            task_alias=metadata["match"].get("taskAlias", {}),
        )

    # def __validate_export_task(self, data, struct):
    #     if not isinstance(data, dict):
    #         return

    #     for key, value in struct.items():
    #         if key not in data:
    #             raise ValueError(f"Missing key: {key}")
    #         value = data[key]
    #         sub_struct = struct[key]
    #         self.__validate_export_task(value, sub_struct)

    def get_export(
        self, export_config: dict, thread_num: int = 16, dsn_cache: bool = False
    ) -> tuple[
        "MolarClient.ExportInfo", "MolarClient.TaskInfo", list["MolarClient.ItemInfo"]
    ]:
        """
        从http客户端导出
        """
        # self.__validate_export_task(export_config, self.__required_structure)
        # 使用 DNS 缓存
        sessions = [requests.Session() for _ in range(thread_num)]
        host = None
        if dsn_cache:
            origin_domain = self.domain
            host = origin_domain.replace("https://", "")
            ip = socket.gethostbyname(host)
            self.domain = f"https://{ip}"
            for session in sessions:
                session.mount("https://", SSLAdapter(server_hostname=host))
        export_info = self.__to_export_info(export_config=export_config)
        task_info = self.get_task_info(
            export_info.task_uid, host=host, session=random.choice(sessions)
        )
        logger.info(f"export: {len(export_info.item_ids)} items")

        items = Parallel(n_jobs=thread_num, backend="threading")(
            delayed(self.process_item)(
                item_id,
                task_info.uid,
                session=random.choice(sessions),
                host=host,
            )
            for item_id in set(export_info.item_ids)
        )
        items = list(filter(None, items))

        if len(items) != len(export_info.item_ids):
            raise ValueError(
                f"items length: {len(items)} != export_info.item_ids length: {len(export_info.item_ids)}"
            )

        if dsn_cache:
            self.domain = origin_domain
        return export_info, task_info, items

    def process_item(
        self,
        item_id: str,
        task_id: str,
        session: requests.Session | None = None,
        host: str | None = None,
    ) -> "MolarClient.ItemInfo":
        """处理条目"""
        item_info = self.get_item_info(item_id=item_id, session=session, host=host)
        labels = self.find_labels(
            task_id=task_id, item_id=item_id, session=session, host=host
        )
        item_info.labels = labels
        return item_info

    def get_task_info(
        self,
        task_id: str,
        session: requests.Session | None = None,
        host: str | None = None,
    ) -> "MolarClient.TaskInfo":
        """
        获取任务信息
        """
        while (
            response_json := get_task_info(
                task_id=task_id,
                token=self.token,
                domain=self.domain,
                session=session,
                host=host,
            )
        ) is None:
            continue
        data: dict = response_json.pop("data")
        return MolarClient.TaskInfo(
            uid=data.pop("_id"),
            domain_id=data.pop("domainId"),
            name=data.pop("name"),
            type=data.pop("type"),
            label_config=data["setting"].pop("labelConfig"),
            label_alias=data["setting"].pop("labelAlias"),
            origin_data=data,
        )

    def get_item_info(
        self,
        item_id: str,
        session: requests.Session | None = None,
        host: str | None = None,
    ) -> "MolarClient.ItemInfo":
        """获取条目"""
        while (
            response_json := get_item_info(
                item_id=item_id,
                token=self.token,
                domain=self.domain,
                session=session,
                host=host,
            )
        ) is None:
            continue
        data = response_json.pop("data")
        return MolarClient.ItemInfo(
            uid=data.pop("_id"),
            task_id=data.pop("taskId"),
            batch_id=data["packageInfo"].pop("_id"),
            info=data.pop("info"),
            origin_data=data,
        )

    def find_labels(
        self,
        task_id: str,
        item_id: str,
        session: requests.Session | None = None,
        host: str | None = None,
    ) -> list["MolarClient.LabelInfo"]:
        """获取标签"""
        while (
            response_json := find_labels(
                task_id=task_id,
                item_id=item_id,
                token=self.token,
                domain=self.domain,
                session=session,
                host=host,
            )
        ) is None:
            continue
        labels: list[dict] = response_json.pop("data")
        return [
            MolarClient.LabelInfo(
                origin_data=label,
                uid=label.pop("_id"),
                task_uid=label.pop("taskId"),
                item_uid=label.pop("itemId"),
                status=label.pop("status"),
                data=label.pop("data"),
            )
            for label in labels
        ]

    @dataclass
    class TaskInfo:
        """任务信息"""

        origin_data: dict
        uid: str
        domain_id: str
        name: str
        type: str
        label_config: dict = field(default_factory=dict)
        label_alias: dict = field(default_factory=dict)
        items: list["MolarClient.ItemInfo"] = field(default_factory=list)

    @dataclass
    class ItemInfo:
        """条目信息"""

        origin_data: list[dict]
        uid: str
        task_id: str
        batch_id: str
        info: dict = field(default_factory=dict)
        labels: list["MolarClient.LabelInfo"] = field(default_factory=list)

    @dataclass
    class LabelInfo:
        """标签信息"""

        origin_data: dict
        uid: str
        task_uid: str
        item_uid: str
        status: str
        data: dict = field(default_factory=dict)

    @dataclass
    class ExportInfo:
        """
        平台导出信息
        """

        origin_data: dict
        task_uid: str
        item_ids: list[str] = field(default_factory=list)
        task_alias: dict = field(default_factory=dict)



import numpy as np
from scipy.spatial.transform import Rotation as R


class Pose:
    x: float
    y: float
    z: float
    qx: float
    qy: float
    qz: float
    qw: float

    def __init__(self, translation, quat):
        self.x, self.y, self.z = translation
        self.qx, self.qy, self.qz, self.qw = quat

    def get_euler(self, seq="xyz", degrees=False):
        return R.from_quat([self.qx, self.qy, self.qz, self.qw]).as_euler(
            seq=seq, degrees=degrees
        )

    @property
    def rot_matrix(self):
        return R.from_quat([self.qx, self.qy, self.qz, self.qw]).as_matrix()

    @property
    def homogeneous_matrix(self):
        """返回齐次矩阵"""
        homogeneous_matrix = np.eye(4)
        homogeneous_matrix[:3, :3] = self.rot_matrix
        homogeneous_matrix[:3, 3] = [self.x, self.y, self.z]
        return homogeneous_matrix

    @property
    def translation(self):
        return np.array([self.x, self.y, self.z])

    @property
    def quaternion(self):
        return np.array([self.qx, self.qy, self.qz, self.qw])

    @staticmethod
    def from_homogeneous_matrix(matrix: np.ndarray) -> "Pose":
        """齐次矩阵构建Pose

        Args:
            matrix (np.ndarray): 齐次矩阵

        Returns:
            Pose: 位姿
        """
        translation = matrix[:3, 3]
        quat = R.from_matrix(matrix[:3, :3]).as_quat()

        return Pose(translation=translation, quat=quat)

    def __matmul__(self, other: "Pose") -> "Pose":
        """
        支持 Pose2 @ Pose2 左乘操作
        等价于: self.homogeneous_matrix @ other.homogeneous_matrix
        """
        if not isinstance(other, Pose):
            raise TypeError("只能将 Pose2 左乘另一个 Pose2")
        new_matrix = self.homogeneous_matrix @ other.homogeneous_matrix
        return Pose.from_homogeneous_matrix(new_matrix)



# /bin/env python
# -*- coding: utf-8 -*-
"""
@author: inklov3
@date: 2025-07-16
@description: molar function tools
"""

import datetime
import hashlib
import json
import sys
import uuid
from pathlib import Path, PurePath
from urllib.parse import unquote, urlparse

from loguru import logger


def file_name(file: Path, suffix: str | None = None) -> str:
    """
    文件名处理
    """
    name = file.name

    if suffix is not None:
        name = name.rsplit(".", 1)[0] + suffix

    return name


def ensure_dir(path: str | Path) -> Path:
    """
    确保目录存在
    """
    if isinstance(path, str):
        path = Path(path)
    path.mkdir(parents=True, exist_ok=True)
    return path


def ensure_file(path: Path | str) -> Path:
    """
    确保父目录存在
    """
    if isinstance(path, str):
        path = Path(path)
    path.parent.mkdir(parents=True, exist_ok=True)
    return path


def init_logger(level: str = "WARNING", logs_path: Path = Path("logs")):
    """
    初始化日志
    """
    if getattr(logger, "_configured", False):
        return

    logger.remove()
    logger.add(
        sink=sys.stdout,
        format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <7}</level> | "
        "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - "
        "<level>{message}</level>",
        level=level,
    )
    if logs_path:
        logs_path.mkdir(parents=True, exist_ok=True)
        log_path = logs_path / f"{datetime.datetime.now():%Y-%m-%d_%H-%M-%S}.log"
        logger.add(
            str(log_path),
            format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <7}</level> | "
            "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - "
            "<level>{message}</level>",
            level=level,
            encoding="utf-8",
        )
    setattr(logger, "_configured", True)


def read_json(path: Path) -> dict:
    """
    读json
    """
    return json.loads(path.read_text())


class NonSerializable:
    """标记字段为不可序列化：序列化时将跳过此字段"""

    def __repr__(self):
        return "<NonSerializable>"


NON_SERIALIZABLE = NonSerializable()


def serialize(obj):
    """
    序列化
    """
    result = None
    # 基础类型或 None
    if isinstance(obj, (str, int, float, bool)) or obj is None:
        result = obj
    # 字典
    elif isinstance(obj, dict):
        result = {k: serialize(v) for k, v in obj.items() if v is not NON_SERIALIZABLE}
    # 列表/元组/集合
    elif isinstance(obj, (list, tuple, set)):
        result = [serialize(v) for v in obj if v is not NON_SERIALIZABLE]
    # 普通对象
    elif hasattr(obj, "__dict__"):
        filtered = {
            k: v for k, v in vars(obj).items() if not isinstance(v, NonSerializable)
        }
        result = serialize(filtered)
    else:
        raise TypeError(f"Unsupported serialization type: {type(obj)}")
    return result


def write_json(data, json_path: Path, indent=2):
    """
    写json
    """
    with open(json_path, "w", encoding="utf8") as file:
        json.dump(data, file, ensure_ascii=False, indent=indent, default=serialize)


def hash_string(s: str, method="md5") -> str:
    """
    字符串hash
    """
    h = hashlib.new(method)
    h.update(s.encode("utf-8"))
    return h.hexdigest()


def partial_hash(file: Path, chunk_size=4 * 1024 * 1024) -> str:
    """
    分块hash
    """
    size = file.stat().st_size
    md5 = hashlib.md5()

    ranges = []
    ranges.append((0, chunk_size))  # 开头

    if size > chunk_size * 2:
        middle = max((size // 2) - (chunk_size // 2), chunk_size)
        ranges.append((middle, chunk_size))  # 中间

    if size > chunk_size:
        ranges.append((size - chunk_size, chunk_size))  # 结尾

    with file.open("rb") as f:
        for offset, length in ranges:
            f.seek(offset)
            md5.update(f.read(length))

    return md5.hexdigest()


def get_uuid():
    """
    获取uuid
    """
    random_uuid = uuid.uuid4()
    return str(random_uuid)


def get_timestamp():
    """
    获取时间戳
    """
    time = datetime.datetime.now()
    return time.strftime("%Y.%m.%d-%H:%M:%S.%f")


def search_file_with_suffix(path: Path, suffixes: list[str]) -> list[Path]:
    """递归查找目录后缀文件"""
    return [f for f in path.rglob(r"*") if f.suffix in suffixes]


def cut_url(
    url: str, start_level: int = 2, end_level: int = -1, with_file_name: bool = False
) -> PurePath:
    """切割url"""
    url_parse = urlparse(url)
    url_path = url_parse.path
    unquoted_url_path = Path(unquote(url_path))
    if with_file_name:
        end_level = len(unquoted_url_path.parts)
    return PurePath(*unquoted_url_path.parts[start_level:end_level])



import datetime
from enum import Enum
from dataclasses import dataclass, field

from .lib import get_uuid

# 分布式任务支持


class TaskStatus(Enum):
    FAILED = -1  # 失败
    SUCCEEDED = 0  # 成功
    PENDING = 1  # 等待
    RUNNING = 2  # 运行
    CANCELLED = 3  # 取消
    RETRYING = 4  # 重试
    TIMEOUT = 5  # 超时


@dataclass
class TaskId:
    tid: str = field(default_factory=get_uuid())
    timestamp: int = field(default=int(datetime.datetime.now().timestamp()))
    status: int = field(default=1)
    args: dict = field(default_factory=dict)
